#include <cmath>
#include <cstddef>
#include <iostream>
#include <numbers>
#include <vector>

#include <filter/filter/low_pass_filter.hpp>

using std::numbers::pi;

namespace Filter {
LowPassFilter::LowPassFilter(std::int32_t size, std::size_t cutoffHz, std::size_t sampleRateHz, const std::vector<float>& window) : IFilter(size) {
  for (std::int32_t i = 0; i < size; i++) {
    int shift = size / 2.0f;
    float sampleTimeS = 1.0f / sampleRateHz;

    if (shift != i) {
      IFilter::impulseResponse[i] = std::sin(2.0f * pi * cutoffHz * sampleTimeS * (i - shift)) / (pi * sampleTimeS * (i - shift));
    } else {
      IFilter::impulseResponse[i] = 2.0f * cutoffHz;
    }

    IFilter::impulseResponse[i] *= sampleTimeS * window[i];
  }
}

} // namespace Filter